require 'matrix'

class TalltyImportExport::ExportPayload
  attr_reader :value, :context

  def initialize(val, header:, &value_handler)
    @value = TalltyImportExport::ExportPayload::Value.new(val)
    @headers = header.items # TalltyImportExport::Attr::ExportHeaderItem array
    @flatten_headers = header.flatten_value
    @value_handler = value_handler
    @context = {
      header_seq_to_value_seq_to_value: {},
      value_inherited_seq_to_header_seq_to_value: {},
      flatten_headers: @flatten_headers,
    }
  end

  def lines
    cell_columns = @headers.map do |header|
      TalltyImportExport::ExportPayload::CellColumn.new(
        header,
        self.value,
        @context,
        &@value_handler
      )
    end

    # 分离的，每个初始 header 为一个矩阵
    matrixes = cell_columns.map do |cell_column|
      cell_column.tallest_cell_cluster.map(&:grow_to_line)
    end

    # 计算所有矩阵中最高高度，用于填充
    max_matrix_height = matrixes.map(&:count).max

    # 填充各个矩阵至等高
    detached_matrixes = matrixes.map do |ragged_col|
      col = ragged_col.uniq
      while col.count < max_matrix_height
        col.push(col[0].count.times.to_a)
      end
      Matrix[*col]
    end

    result = detached_matrixes.reduce do |out, matrix|
      Matrix.hstack(out, matrix)
    end.to_a

    # result.map { |x| p x.map { |xx| xx.try(:value)}}
    result
  end
end

class TalltyImportExport::ExportPayload::Value
  attr_reader :value, :seq_chain#, :inherited_seq

  def initialize(val, seq_chain=[], inherited_seq: false)
    @value = val
    @seq_chain = inherited_seq ? seq_chain : [*seq_chain, seq]
  end

  def seq
    @seq ||= SecureRandom.uuid
  end
end

# 一个 竖列
class TalltyImportExport::ExportPayload::CellColumn
  attr_reader :header, :payload, :context, :cell_clusters, :flatten_cells, :tallest_cell_cluster

  def initialize header, value, context, &value_handler
    @header = header
    @value = value
    @context = context
    @cell_clusters = []
    @value_handler = value_handler

    if header.children
      @cell_clusters = @value.value[header.key.to_sym].map do |val|
        TalltyImportExport::ExportPayload::Cell.new(
          @header,
          TalltyImportExport::ExportPayload::Value.new(val, @value.seq_chain),
          @context,
          &@value_handler
        )
      end
    else
      @cell_clusters = [
        TalltyImportExport::ExportPayload::Cell.new(
          @header,
          TalltyImportExport::ExportPayload::Value.new(
            @value_handler.call(@value.value, @header.as_json.symbolize_keys),
            @value.seq_chain,
          ),
          @context,
          &@value_handler
        )
      ]
    end

    @flatten_cells = []

    cell_divise
  end

  def cell_divise
    @flatten_cells = @cell_clusters.map do |cell_cluster|
      cell_cluster.divise.cell_cluster
    end.flatten

    depth_to_cells_mapping = @flatten_cells.group_by { |cell| cell.depth }

    @tallest_cell_cluster = depth_to_cells_mapping.values.max { |a, b| a.count <=> b.count }

    self
  end
end

class TalltyImportExport::ExportPayload::Cell
  attr_accessor :cell_cluster, :val, :header, :parent_path

  # cell_cluster 是表格纵向 集合
  # cell_the_same_line 是表格横向 集合
  def initialize header, value, context, parent_path=[], &value_handler
    @value = value
    @header = header
    @value_handler = value_handler

    # context
    @context = context

    value.seq_chain.each do |s|
    # s = value.seq
      unless (@context[:header_seq_to_value_seq_to_value][header.seq])
        @context[:header_seq_to_value_seq_to_value][header.seq] = {}
      end
      @context[:header_seq_to_value_seq_to_value][header.seq][s] = value # NOTE: value 对象
    end

    # unless (@context[:value_inherited_seq_to_header_seq_to_value][value.inherited_seq])
    #   @context[:value_inherited_seq_to_header_seq_to_value][value.inherited_seq] = {}
    # end
    # @context[:value_inherited_seq_to_header_seq_to_value][value.inherited_seq][header.seq] = value # NOTE: value 对象

    # parent
    @parent_path = parent_path

    @cell_cluster = []
  end

  def seq
    @seq ||= SecureRandom.uuid
  end

  def depth
    @parent_path.count
  end

  def header_chain_seqs
    [*@parent_path, self].map { |cell| cell.header.cluster_seqs }.reduce(:concat)
  end

  def divise
    if @header.children.present?
      new_cells = @header.children.map do |child_header|
        if (child_header.children.present?)
          val = @value.value[child_header.key.to_sym]
          val.map do |val|
            TalltyImportExport::ExportPayload::Cell.new(
              child_header,
              TalltyImportExport::ExportPayload::Value.new(val, @value.seq_chain, inherited_seq: true),
              @context,
              [*self.parent_path, self],
              &@value_handler
            )
          end
        else
          val = @value_handler.call(@value.value, child_header.as_json.symbolize_keys)
          TalltyImportExport::ExportPayload::Cell.new(
            child_header,
            TalltyImportExport::ExportPayload::Value.new(val, @value.seq_chain),
            @context,
            [*self.parent_path, self],
            &@value_handler
          )
        end
      end.flatten
      new_cells.each(&:divise)
      @cell_cluster.concat(new_cells.map(&:cell_cluster))
    else
      @cell_cluster = [self]
    end

    self
  end

  def grow_to_line
    @context[:flatten_headers].
      select do |header|
        header_chain_seqs.include?(header.seq)
      end.
      map do |header|
        target_value = nil

        # if @context[:value_inherited_seq_to_header_seq_to_value][@value.inherited_seq]&.[](header.seq)
        #   target_value = @context[:value_inherited_seq_to_header_seq_to_value][@value.inherited_seq][header.seq]
        # end

        @value.seq_chain.uniq.reverse.each do |seq|
          next if target_value
          # seq = @value.seq
          if @context[:header_seq_to_value_seq_to_value][header.seq]&.[](seq)
            target_value = @context[:header_seq_to_value_seq_to_value][header.seq][seq]
          end
      end

# require 'irb'
# if target_value == 'c1_4____'
# binding.irb

# end
      # @value.seq_chain.uniq.reverse.each do |seq|
      #   next if target_value
      #   if @context[:header_seq_to_value_seq_to_value][seq]&.[](header.seq)
      #     target_value = @context[:header_seq_to_value_seq_to_value][seq][header.seq]
      #   end
      # end


      target_value
    end
  end
end
